#!/usr/bin/python
import sys
import signal
import string
import traceback
import time
import re
import socket
import os
import commands
import thread
import logging
global producer
global log
global debug

def createDaemon(err_file,log_file,log_level='ERROR'):
    global log
    global debug

    if not debug:
        try:
            pid = os.fork()
        except OSError, e:
            return((e.errno, e.strerror))

        if (pid == 0):
            os.setsid()
            signal.signal(signal.SIGHUP, signal.SIG_IGN)
            try:
                pid = os.fork()
            except OSError, e:
                return((e.errno, e.strerror))
            if (pid == 0):
                os.umask(022)
            else:
                os._exit(0)
        else:
            os._exit(0)
        try:
            maxfd=os.sysconf("SC_OPEN_MAX")
        except (AttributeError, ValueError):
            maxfd=256
        for fd in range(3, maxfd):
            try:
                os.close(fd)
            except OSError:
                pass
        os.close(0)
        os.open("/dev/null", os.O_RDONLY)
        os.close(1)
        os.open("/dev/null", os.O_WRONLY)

        #
        # connect stderr to error logfile
        #

        e = os.open(err_file, os.O_WRONLY | os.O_APPEND | os.O_CREAT, 0644)
        os.dup2(e, 2)
        os.close(e)
        sys.stderr = os.fdopen(2, 'a', 0)
        sys.stderr.write("\nstarted at " + time.ctime() + "\n")
        os.chdir("/")

    log = logging.getLogger('lcg-mon-gridftp')
    hdlr = logging.FileHandler(log_file)
    formatter = logging.Formatter('%(asctime)s: [%(levelname)s] %(message)s')
    hdlr.setFormatter(formatter)
    log.addHandler(hdlr)
    log.setLevel(logging.__dict__.get(log_level,'ERROR'))

def createproducer(table):
    global producer

    while 1:
	if producer != None:
	    try:
		producer.close()
	    except (rgma.RGMAException, rgma.RemoteException, rgma.UnknownResourceException),e:
		log.error("R-GMA Error closing old primary producer.")
	    except(),e:
		log.error("Error closing old primary producer.")
	    producer = None
        try:
            storage = rgma.Storage(rgma.StorageType.MEMORY)
            properties = rgma.ProducerProperties(storage)
            terminationInterval = rgma.TimeInterval(60,rgma.Units.MINUTES)
            producer = rgma.PrimaryProducer(terminationInterval,properties,[])
            predicate = ""
            latestRetentionPeriod = rgma.TimeInterval(30, rgma.Units.MINUTES)
            historyRetentionPeriod = rgma.TimeInterval(30, rgma.Units.MINUTES)
            producer.declareTable(table,predicate,latestRetentionPeriod,
                                  historyRetentionPeriod)
            log.info("Creating primary producer successful.")
            return
        except Exception, e:
            log.error("Error creating primary producer.")
            log.error(e)
            log.error("Will try again in 60s")
            time.sleep(60)


def handler(signum, frame):
    global producer
    log.info("Shutting Down PrimaryProducer")
    producer.close()
    sys.exit(0)

def readfile(offset, filename):
    try:
	logfile=open(filename, 'r')
    except IOError, (e, s):
        log.error("Cannot open file %s: %s" % (filename, s))
        return []
    except:
        log.error("Cannot open file %s" % filename)
        return []

    logfile.seek(offset, 0)
    entries=logfile.readlines()
    logfile.close()
    return entries

def main(filename, table):
    #Set up Signal handler for SIGTERM (kill -15) and SIGINT (ctrl-c)
    signal.signal(signal.SIGTERM, handler)
    signal.signal(signal.SIGINT, handler)

    #Name of the file in which the logfile offset is remembered
    prev_file_offset = sys.argv[1].replace('etc', 'var/tmp')
    prev_file_offset = prev_file_offset.replace('.conf', '.tmp')

    os.system("touch " + prev_file_offset)
    last_offset = open(prev_file_offset, 'r+')
    line = last_offset.readline()
    line = re.sub("\s.*", "", line)
    offset = int("0" + line)

    #log rotated file is
    original_file = filename
    rotated_file = filename + ".1"

    while 1:
        log.debug("while, main")
        #If we have just used the rotated file, swap back.
        if filename == rotated_file:
            filename = original_file
            offset = 0

	while 1:
            log.debug("while, check log file")
	    try:
		file_size = os.stat(filename).st_size

		if file_size > offset:
		    break

		if file_size < offset:
                    # assume logfile got rotated, new file created
                    # so try the log rotated file.
                    filename=rotated_file
                    break
            except:
		# assume logfile got rotated, new file not yet created
                pass

            time.sleep(30)

        entries = readfile(offset, filename)
        if len(entries)==0:
            time.sleep(30)
            continue

        for entry in entries:
            log.debug("Next Entry")
            if entry[-1] != "\n":
                time.sleep(10)
                break

	    command = parser.parse_entry(entry)
            # update the remembered offset per entry
            offset=offset + len(entry)
            last_offset.seek(0)
            last_offset.write("%-20u\n" % offset)
            last_offset.flush()

            #Ingnore bad lines
            if (command == "ERROR"):
                log.error("Bad Entry")
                log.error(entry)
                continue

            if (command == "Ignore"):
                continue

            while 1:
                log.debug("while, insert")
                log.debug(command)
                try:
                    producer.insert(command)
                    break
                except rgma.RGMAUserException, e:
                    log.error("Bad user input.")
                    log.error(e)
                    log.error("Ignoring Entry.")
                    log.error(command)
                    break
                except (rgma.RemoteException, rgma.RGMABufferFullException),e:
                    log.error("Failed to insert tuple.")
                    log.error(e)
                    log.error("Will try again in 60s")
                    time.sleep(60)
                except (rgma.RGMAException, rgma.UnknownResourceException),e:
                    log.error("Failed to insert tuple.")
                    log.error(e)
                    log.error("Will try and create a new producer")
                    createproducer(table)
                except Exception, e:
                    log.error("Failed to insert tuple.")
                    log.error(e)
                    log.error("Ignoring Entry.")
                    log.error(command)
                    break

if __name__ == '__main__':
    global producer
    filename=sys.argv[1]
    config_file=open(filename, 'r')
    params={}

    for line in config_file:
	if line and line[0]!="#":
	    if len(line.split('=')) == 2:
		key, value = line.split('=')
		params[key.strip()] = value.strip()

    config_file.close()
    log_file=sys.argv[1].replace('etc', 'var/log')
    log_file=log_file.replace('.conf', '.log')
    err_file=sys.argv[1].replace('etc', 'var')
    err_file=err_file.replace('.conf', '.err')
    open(log_file, 'a').close()
    open(err_file, 'a').close()
    debug = params.get('debug','0') != '0'

    path=sys.argv[0].replace('libexec/lcg-mon-logfile-daemon.py', 'lib/python')
    sys.path.append(path)
    parser =  __import__(params['PARSER'])

    createDaemon(err_file,log_file,params.get('log_level','ERROR'))

    sys.path.append(params['RGMA_HOME'] + '/lib/python')
    os.environ['RGMA_HOME']=params['RGMA_HOME']
    import rgma
    producer = None
    createproducer(params["TABLE"])
    main(params["LOG_FILE"], params["TABLE"])

